// file      : xsd/cxx/parser/parser-source.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2007 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/parser-source.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      struct Enumeration: Traversal::Enumeration,
                          protected virtual Context
      {
        Enumeration (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String const& ret (ret_type (e));

          SemanticGraph::Type& base (e.inherits ().base ());

          Boolean same (ret == ret_type (base));

          if (same || ret == L"void")
          {
            String const& name (ename (e));

            os << "// " << name << endl
               << "//" << endl
               << endl;

            os << ret << " " << name << "::" << endl
               << post_name (e) << " ()"
               << "{";

            if (same)
              os << "return " << post_name (base) << " ();";

            os << "}";
          }
        }
      };

      //
      //
      struct List: Traversal::List, protected virtual Context
      {
        List (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          String const& name (ename (l));
          SemanticGraph::Type& t (l.argumented ().type ());

          String item (unclash (name, "item"));

          os << "// " << name << endl
             << "//" << endl
             << endl;

          // item
          //
          String const& arg (arg_type (t));

          os << "void " << name << "::" << endl
             << item;

          if (arg == L"void")
            os << " ()";
          else
            os << " (" << arg << ")";

          os << "{"
             << "}";

          // post
          //
          if (ret_type (l) == L"void")
            os << "void " << name << "::" << endl
               << post_name (l) << " ()"
               << "{"
               << "}";

          // parse_item
          //
          String inst (L"_xsd_" + item + L"_");
          String const& post (post_name (t));

          os << "void " << name << "::" << endl
             << "_xsd_parse_item (const " << string_type << "& v)"
             << "{"
             << "if (this->" << inst << ")"
             << "{"
             << "this->" << inst << "->pre ();"
             << "this->" << inst << "->_pre_impl ();"
             << "this->" << inst << "->_characters (v);"
             << "this->" << inst << "->_post_impl ();";

          if (ret_type (t) == L"void")
            os << "this->" << inst << "->" << post << " ();"
               << "this->" << item << " ();";
          else
            os << "this->" << item << " (this->" << inst << "->" <<
              post << " ());";

          os << "}"
             << "}";
        }
      };

      //
      //
      struct Union: Traversal::Union, protected virtual Context
      {
        Union (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          if (ret_type (u) == L"void")
          {
            String const& name (ename (u));

            os << "// " << name << endl
               << "//" << endl
               << endl;

            os << "void " << name << "::" << endl
               << post_name (u) << " ()"
               << "{"
               << "}";
          }
        }
      };

      //
      //
      struct StartElement : Traversal::Element,
                            protected virtual Context
      {
        StartElement (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
	  if (skip (e)) return;

          String const& name (ename (e));
          String const& inst (emember (e));

          if (e.qualified ())
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && " <<
              "ns == " << L << "\"" << e.namespace_ ().name () << "\")"
               << "{";
          }
          else
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && ns.empty ())"
               << "{";
          }

          os << "this->" << complex_base << "::context_.top ().parser_ = " <<
            "this->" << inst << ";"
             << endl
             << "if (this->" << inst << ")" << endl
             << "this->" << inst << "->pre ();" // _start_element calls _pre
             << endl
             << "return true;"
             << "}";
        }
      };


      //
      //
      struct EndElement : Traversal::Element,
                          protected virtual Context
      {
        EndElement (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
	  if (skip (e)) return;

          String const& name (ename (e));
          String const& inst (emember (e));

          if (e.qualified ())
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && " <<
              "ns == " << L << "\"" << e.namespace_ ().name () << "\")"
               << "{";
          }
          else
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && ns.empty ())"
               << "{";
          }

          // _end_element calls post
          //

          SemanticGraph::Type& type (e.type ());
          String const& post (post_name (type));

          os << "if (this->" << inst << ")"
             << "{";

          if (ret_type (type) == L"void")
            os << "this->" << inst << "->" << post << " ();"
               << "this->" << name << " ();";
          else
            os << "this->" << name << " (this->" << inst << "->" <<
              post << " ());";

          os << "}"
             << "return true;"
             << "}";
        }
      };

      //
      //
      struct Attribute : Traversal::Attribute,
                         protected virtual Context
      {
        Attribute (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String const& name (ename (a));
          String const& inst (emember (a));

          if (a.qualified ())
          {
            os << "if (n == " << L << "\"" << a.name () << "\" && " <<
              "ns == " << L << "\"" << a.namespace_ ().name () << "\")"
               << "{";
          }
          else
          {
            os << "if (n == " << L << "\"" << a.name () << "\" && ns.empty ())"
               << "{";
          }

          SemanticGraph::Type& type (a.type ());
          String const& post (post_name (type));
          String const& ret (ret_type (type));

          os << "if (this->" << inst << ")"
             << "{"
             << "this->" << inst << "->pre ();"
             << "this->" << inst << "->_pre_impl ();"
             << "this->" << inst << "->_characters (v);"
             << "this->" << inst << "->_post_impl ();";

          if (ret == L"void")
            os << "this->" << inst << "->" << post << " ();"
               << "this->" << name << " ();";
          else
            os << "this->" << name << " (this->" << inst << "->" <<
              post << " ());";

          os << "}"
             << "return true;"
             << "}";
        }
      };

      //
      //
      struct ParserHook: Traversal::Member,
                         protected virtual Context
      {
        ParserHook (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& arg (arg_type (m.type ()));

          os << "void " << ename (m.scope ()) << "::" << endl
             << ename (m);

          if (arg == L"void")
            os << " ()";
          else
            os << " (" << arg << ")";

          os << "{"
             << "}";
        }
      };

      //
      //
      struct Complex : Traversal::Complex,
                       protected virtual Context
      {
        Complex (Context& c)
            : Context (c),
              parser_hook_ (c),
              start_element_ (c),
              end_element_ (c),
              attribute_ (c)
        {
          names_parser_hook_ >> parser_hook_;
          names_start_element_ >> start_element_;
          names_end_element_ >> end_element_;
          names_attribute_ >> attribute_;
        }

        virtual Void
        traverse (Type& c)
        {
          Boolean he (has<Traversal::Element> (c));
          Boolean ha (has<Traversal::Attribute> (c));

          String const& ret (ret_type (c));
          Boolean same (c.inherits_p () &&
                        ret == ret_type (c.inherits ().base ()));

          if (!(he || ha || same || ret == L"void"))
            return;

          String const& name (ename (c));
          Boolean restriction (restriction_p (c));

          os << "// " << name << endl
             << "//" << endl
             << endl;

          // Parser hooks.
          //
          if (!restriction)
            names (c, names_parser_hook_);

          if (same || ret == L"void")
          {
            os << ret << " " << name << "::" << endl
               << post_name (c) << " ()"
               << "{";

            if (same)
              os << "return " << post_name (c.inherits ().base ()) << " ();";

            os << "}";
          }

          // The rest is parsing/validation code which is generated in
          // validation-source.cxx.
          //
          if (validation && restriction)
            return;


          // _start_element_impl & _end_element_impl
          //
          if (he && !validation)
          {
            os << "bool " << name << "::" << endl
               << "_start_element_impl (const " << string_type << "& ns," << endl
               << "const " << string_type << "& n)"
               << "{"
               << "if (this->";

            if (c.inherits_p ())
              os << fq_name (c.inherits ().base ());
            else
              os << complex_base;

            os << "::_start_element_impl (ns, n))"
               << "{"
               << "return true;"
               << "}";

            names (c, names_start_element_);

            os << "return false;"
               << "}";


            // _end_element_impl
            //
            os << "bool " << name << "::" << endl
               << "_end_element_impl (const " << string_type << "& ns," << endl
               << "const " << string_type << "& n)"
               << "{"
               << "if (this->";

            if (c.inherits_p ())
              os << fq_name (c.inherits ().base ());
            else
              os << complex_base;

            os << "::_end_element_impl (ns, n))"
               << "{"
               << "return true;"
               << "}";

            names (c, names_end_element_);

            os << "return false;"
               << "}";
          }


          if (ha && !validation)
          {
            // _attribute_impl
            //
            os << "bool " << name << "::" << endl
               << "_attribute_impl (const " << string_type << "& ns," << endl
               << "const " << string_type << "& n," << endl
               << "const " << string_type << "& v)"
               << "{"
               << "if (this->";

            if (c.inherits_p ())
              os << fq_name (c.inherits ().base ());
            else
              os << complex_base;

            os << "::_attribute_impl (ns, n, v))"
               << "{"
               << "return true;"
               << "}";

            names (c, names_attribute_);

            os << "return false;"
               << "}";
          }
        }

      private:
        //
        //
        ParserHook parser_hook_;
        Traversal::Names names_parser_hook_;

        //
        //
        StartElement start_element_;
        Traversal::Names names_start_element_;

        //
        //
        EndElement end_element_;
        Traversal::Names names_end_element_;

        //
        //
        Attribute attribute_;
        Traversal::Names names_attribute_;
      };
    }

    Void
    generate_parser_source (Context& ctx)
    {
      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names schema_names;

      Namespace ns (ctx);
      Traversal::Names names;

      schema >> sources >> schema;
      schema >> schema_names >> ns >> names;

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;

      schema.dispatch (ctx.schema_root);
    }
  }
}

