// file      : processing/restriction/processor.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2006-2007 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <processing/restriction/processor.hxx>

#include <elements.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <cult/containers/set.hxx>

#include <iostream>
using std::wcerr;
using std::endl;

namespace Processing
{
  using namespace Cult;

  namespace SemanticGraph = XSDFrontend::SemanticGraph;
  namespace Traversal = XSDFrontend::Traversal;

  typedef WideString String;

  namespace Restriction
  {
    namespace
    {
      //
      //
      struct Complex: Traversal::Complex
      {
        Complex (SemanticGraph::Schema& schema)
            : schema_ (schema)
        {
        }

        virtual Void
        traverse (Type& c)
        {
          using namespace SemanticGraph;
          using SemanticGraph::Complex;

          if (c.context ().count ("processing-restriction-seen"))
            return;

          c.context ().set ("processing-restriction-seen", true);

          // Current implementation of semantic graph uses the same Restricts
          // edge for both simple type/content restriction and complex content
          // restriction. Here we are interested in the complex content only.
          //
          //
          if (c.inherits_p () && c.inherits ().is_a<Restricts> ())
          {
            // Go down our inheritance hierarchy until the end or the next
            // restriction.
            //
            Complex* pbase (&c);

            while (pbase = dynamic_cast<Complex*> (
                     &pbase->inherits ().base ()))
            {
              traverse (*pbase); // Make sure our base is processed.
              merge_attributes (c, *pbase);

              if (!pbase->inherits_p () ||
                  pbase->inherits ().is_a<Restricts> ())
                break;
            }
          }

          // Traverse anonymous types (via elements & attributes).
          //
          Traversal::Complex::names (c);
        }

      private:
        Void
        merge_attributes (SemanticGraph::Complex& c,
                          SemanticGraph::Complex& base)
        {
          using namespace SemanticGraph;

          for (Scope::NamesIterator i (base.names_begin ()),
                 e (base.names_end ()); i != e; ++i)
          {
            Attribute* prot (dynamic_cast<Attribute*> (&i->named ()));

            if (prot == 0)
              continue;

            Boolean found (false);
            Name name (prot->name ());
            Scope::NamesIteratorPair r (c.find (name));

            for (; r.first != r.second; ++r.first)
            {
              Attribute* a (dynamic_cast<Attribute*> (&r.first->named ()));

              if (a == 0)
                continue;

              if (prot->qualified ())
              {
                if (a->qualified () &&
                    prot->namespace_ ().name () == a->namespace_ ().name ())
                {
                  found = true;
                  break;
                }
              }
              else
              {
                if (!a->qualified ())
                {
                  found = true;
                  break;
                }
              }
            }

            if (!found)
            {
              Attribute& a (
                schema_.new_node<Attribute> (prot->file (),
                                             prot->line (),
                                             prot->column (),
                                             prot->optional (),
                                             prot->global (),
                                             prot->qualified ()));

              schema_.new_edge<Names> (c, a, name);

              if (prot->qualified ())
                schema_.new_edge<BelongsToNamespace> (a, prot->namespace_ ());

              // Transfer default and fixed values if any.
              //
              if (prot->fixed ())
                a.fixed (prot->value ());
              else if (prot->default_ ())
                a.default_ (prot->value ());

              // Belongs edge.
              //
              schema_.new_edge<Belongs> (a, prot->type ());
            }
          }
        }

      private:
        SemanticGraph::Schema& schema_;
      };

      // Go into implied/included/imported schemas while making sure
      // we don't process the same stuff more than once.
      //
      struct Uses: Traversal::Uses
      {
        virtual Void
        traverse (Type& u)
        {
          SemanticGraph::Schema& s (u.schema ());

          if (!s.context ().count ("processing-restriction-seen"))
          {
            s.context ().set ("processing-restriction-seen", true);
            Traversal::Uses::traverse (u);
          }
        }
      };
    }

    Void Processor::
    process (SemanticGraph::Schema& tu, SemanticGraph::Path const& file)
    {
      Traversal::Schema schema;
      Uses uses;

      schema >> uses >> schema;

      Traversal::Names schema_names;
      Traversal::Namespace ns;
      Traversal::Names ns_names;

      schema >> schema_names >> ns >> ns_names;

      Complex complex_type (tu);
      AnonymousBase anonymous (complex_type);

      ns_names >> complex_type;
      ns_names >> anonymous;

      Traversal::Names names;

      complex_type >> names >> anonymous;

      // Some twisted schemas do recusive inclusions.
      //
      tu.context ().set ("processing-restriction-seen", true);

      schema.dispatch (tu);
    }
  }
}
